/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "maya/maya.pmh"

#include "maya/MFnSingleIndexedComponent.h"

#define FE_SAM_DETAIL_DEBUG		FALSE

namespace fe
{

I32 SurfaceAccessorMaya::createBlindDataID(String a_attrName,String a_attrType)
{
	const String attrKey=a_attrName+":"+a_attrType;
	I32& rBlindDataID=m_spMaster->catalog()->catalog<I32>(
			"MayaBlindID",attrKey,-1);

	if(rBlindDataID>=0)
	{
		//* created already
		return rBlindDataID;
	}

#if FE_SAM_BLIND_DEBUG
	feLog("SurfaceAccessorMaya::createBlindDataID"
			" creating blind data for \"%s\"\n",attrKey.c_str());
#endif

	const String searchName=a_attrName+(a_attrType=="SpatialVector"? "_x": "");

	rBlindDataID=0x0;
	while(m_pFnMesh->isBlindDataTypeUsed(rBlindDataID))
	{
		//* search for already existing match

		MStringArray longNames;
		MStringArray shortNames;
		MStringArray formatNames;
		MStatus status=m_pFnMesh->getBlindDataAttrNames(rBlindDataID,
				longNames,shortNames,formatNames);

#if FE_SAM_BLIND_DEBUG
		feLog("looking for \"%s\" blind %d long \"%s\""
				" short \"%s\" format \"%s\"\n",
				searchName.c_str(),
				rBlindDataID,longNames[0].asChar(),
				shortNames[0].asChar(),formatNames[0].asChar());
#endif

		if(searchName==longNames[0].asChar())
		{
			return rBlindDataID;
		}

		rBlindDataID++;
	}

#if FE_SAM_BLIND_DEBUG
	feLog("SurfaceAccessorMaya::createBlindDataID"
			" blindDataID %d\n",rBlindDataID);
#endif

	MStringArray longNames;
	MStringArray shortNames;
	MStringArray formatNames;

	// format names: int float double boolean string binary

	if(a_attrType=="SpatialVector")
	{
		String name=a_attrName+"_x";
		longNames.append(name.c_str());
		shortNames.append(name.c_str());
		formatNames.append("double");

		name=a_attrName+"_y";
		longNames.append(name.c_str());
		shortNames.append(name.c_str());
		formatNames.append("double");

		name=a_attrName+"_z";
		longNames.append(name.c_str());
		shortNames.append(name.c_str());
		formatNames.append("double");
	}
	else
	{
		longNames.append(a_attrName.c_str());
		shortNames.append(a_attrName.c_str());

		if(a_attrType=="Real")
		{
			formatNames.append("double");
		}
		else if(a_attrType=="I32")
		{
			formatNames.append("int");
		}
		else if(a_attrType=="String")
		{
			formatNames.append("string");
		}
	}

	MStatus status=m_pFnMesh->createBlindDataType(
			rBlindDataID,longNames,shortNames,formatNames);
	if(!status)
	{
		feLog("SurfaceAccessorMaya::createBlindDataID type \"%s\" name \"%s\""
				" createBlindDataType %p failed\n",
				a_attrType.c_str(),a_attrName.c_str(),rBlindDataID);
	}

	return rBlindDataID;
}

void SurfaceAccessorMaya::updateGroupList(void)
{
	m_groupList.clear();
	if(m_attrName.empty())
	{
		return;
	}
	if(m_element!=SurfaceAccessibleI::e_pointGroup &&
			m_element!=SurfaceAccessibleI::e_primitiveGroup)
	{
		return;
	}

	SurfaceAccessibleI::Element storeElement=m_element;

	//* temporary switch to use common methods
	if(m_element==SurfaceAccessibleI::e_pointGroup)
	{
		m_element=SurfaceAccessibleI::e_point;
	}
	else
	{
		m_element=SurfaceAccessibleI::e_primitive;
	}

	//* TODO scan all points/primitives
	//* if group attribute is non-zero, add to group list

	delete[] m_pIntegerCache;
	m_pIntegerCache=NULL;

	const U32 elements=count();
	for(U32 index=0;index<elements;index++)
	{
		if(integer(index)>0)
		{
			m_groupList.push_back(index);
		}
	}

	//* restore original
	m_element=storeElement;

	delete[] m_pIntegerCache;
	m_pIntegerCache=NULL;

}

//* static
String SurfaceAccessorMaya::readDetailAt(MDataHandle a_meshData,I32 a_id)
{
	MFnGeometryData geoData(a_meshData.asMesh());
	MStatus status;

	MObject singles=geoData.objectGroupComponent(a_id,&status);
	if(!status)
	{
		feLog("SurfaceAccessorMaya::readDetailAt"
				" failed to get components at %d because \"%s\"\n",
				a_id,status.errorString().asChar());
	}

	MFnSingleIndexedComponent singleComp(singles);
	MIntArray elements;
	status=singleComp.getElements(elements);

	const int len=elements.length();

	char buffer[len];

	for(U32 m=0;m<len;m++)
	{
		const U32 verifyIndex=(elements[m]>>8);
		if(verifyIndex!=m)
		{
			return "";
		}

		buffer[m]=(elements[m]&0xFF);

//		feLog("read %d:%d %p %d:%d '%c'\n",m,verifyIndex,elements[m],
//				elements[m]&(~0xFF),buffer[m],buffer[m]? buffer[m]: ' ');
	}

	return buffer;
}

//* static
void SurfaceAccessorMaya::writeDetailAt(MDataHandle a_meshData,I32 a_id,
	const String& a_rString)
{
	MObject meshObject=a_meshData.asMesh();
	MFnGeometryData geoData(meshObject);
	MStatus status;

	if(!geoData.hasObjectGroup(a_id))
	{
		status=geoData.addObjectGroup(a_id);
		if(!status)
		{
			feLog("SurfaceAccessorMaya::writeDetailAt"
					"failed to add at %d because \"%s\"\n",
					a_id,status.errorString().asChar());
		}
	}

	MFnSingleIndexedComponent singleComp;
	MObject singles=singleComp.create(MFn::kMeshVertComponent,&status);
	if(!status)
	{
		feLog("SurfaceAccessorMaya::writeDetailAt"
				"failed to create components at %d because \"%s\"\n",
				a_id,status.errorString().asChar());
	}

	const char* buffer=a_rString.c_str();
	const I32 len=a_rString.length();
	for(I32 m=0;m<len;m++)
	{
		const I32 code=(m<<8) + buffer[m];
		singleComp.addElement(code);

//		feLog("write %d %p\n",m,code);
	}
	singleComp.addElement(len<<8);

	status=geoData.setObjectGroupComponent(a_id,singles);
	if(!status)
	{
		feLog("SurfaceAccessorMaya::writeDetailAt"
				"failed to add components because \"%s\"\n",
				status.errorString().asChar());
	}

	//* NOTE needed to propagate changes
	a_meshData.set(meshObject);
}

//* static
void SurfaceAccessorMaya::removeDetailAt(MDataHandle a_meshData,I32 a_id)
{
	MObject meshObject=a_meshData.asMesh();
	MFnGeometryData geoData(meshObject);
	MStatus status;

	if(!geoData.hasObjectGroup(a_id))
	{
		status=geoData.removeObjectGroup(a_id);
		if(!status)
		{
			feLog("SurfaceAccessorMaya::removeDetailAt"
					"failed to remove at %d because \"%s\"\n",
					a_id,status.errorString().asChar());
		}
	}

	//* NOTE needed to propagate changes
	a_meshData.set(meshObject);
}

//* static
I32 SurfaceAccessorMaya::findDetailKey(MDataHandle a_meshData,
	const String& a_rKey,BWORD a_createMissing)
{
	MFnGeometryData geoData(a_meshData.asMesh());
	MStatus status;

	const int groupCount=geoData.objectGroupCount(&status);
	if(!status)
	{
		feLog("SurfaceAccessorMaya::findDetailKey"
				" failed to get group count because \"%s\"\n",
				status.errorString().asChar());
	}
	I32 maxGroup= -1;
	for(U32 groupIndex=0;groupIndex<groupCount;groupIndex++)
	{
		const int group=geoData.objectGroup(groupIndex,&status);
		if(!(groupIndex%1) && readDetailAt(a_meshData,group)==a_rKey)
		{
#if FE_SAM_DETAIL_DEBUG
			feLog("SurfaceAccessorMaya::findDetailKey found %d \"%s\"\n",
					group,a_rKey.c_str());
#endif
			return group;
		}
		if(maxGroup<group)
		{
			maxGroup=group;
		}
	}
	maxGroup++;

	if(!a_createMissing)
	{
		return -1;
	}

#if FE_SAM_DETAIL_DEBUG
	feLog("SurfaceAccessorMaya::findDetailKey create %d \"%s\"\n",
			maxGroup,a_rKey.c_str());
#endif

	writeDetailAt(a_meshData,maxGroup,a_rKey);
	return maxGroup;
}

//* static
void SurfaceAccessorMaya::dumpDetail(MDataHandle a_meshData)
{
	MFnGeometryData geoData(a_meshData.asMesh());
	MStatus status;

	const int groupCount=geoData.objectGroupCount(&status);
	if(!status)
	{
		feLog("SurfaceAccessorMaya::dumpDetail"
				" failed to get group count because \"%s\"\n",
				status.errorString().asChar());
	}
	I32 maxGroup=0;
	for(U32 groupIndex=0;groupIndex<groupCount;groupIndex++)
	{
		const int group=geoData.objectGroup(groupIndex,&status);
		feLog("%d %d \"%s\"\n",groupIndex,group,
				readDetailAt(a_meshData,group).c_str())
	}
}

void SurfaceAccessorMaya::setDetail(const String& a_rKey,
	const String& a_rValue)
{
	const I32 keyId=findDetailKey(m_meshData,a_rKey,TRUE);
	writeDetailAt(m_meshData,keyId+1,a_rValue);

#if FE_SAM_DETAIL_DEBUG
	feLog("SurfaceAccessorMaya::setDetail %d \"%s\" to \"%s\"\n",
			keyId,a_rKey.c_str(),a_rValue.c_str());
	dumpDetail(m_meshData);
#endif
}

String SurfaceAccessorMaya::getDetail(const String& a_rKey)
{
	const I32 keyId=findDetailKey(m_meshData,a_rKey,FALSE);
	if(keyId<0)
	{
		return "";
	}

	String text=readDetailAt(m_meshData,keyId+1);

#if FE_SAM_DETAIL_DEBUG
	feLog("SurfaceAccessorMaya::getDetail %d \"%s\" as \"%s\"\n",
			keyId,a_rKey.c_str(),text.c_str());
#endif

	return text;
}

} // namespace
