/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SBO_DEBUG	FALSE
#define FE_SBO_VERBOSE	FALSE

#include "operator/operator.pmh"

using namespace fe;

void SurfaceBindOp::initialize(void)
{
	catalog<String>("BindPrefix")="bind";
	catalog<String>("BindPrefix","label")="Bind Attribute Prefix";
	catalog<bool>("BindPrefix","joined")=true;
	catalog<String>("BindPrefix","hint")=
			"Prefix of attribute names added to output surface describing"
			" where each point on the input surface is attached to"
			" the associated driver";

	catalog<String>("DriverGroup")="";
	catalog<String>("DriverGroup","label")="Driver Group";
	catalog<String>("DriverGroup","hint")=
			"Primitive Group in Driver Surface that encompass the binding."
			"  Primitives outside this group will not be"
			" considered during binding."
			"  If no group is specified, the entire driver is made available.";

	catalog<bool>("Bind Fragments")=false;
	catalog<bool>("Bind Fragments","joined")=true;
	catalog<String>("Bind Fragments","hint")=
			"Bind collections of points or primitives with"
			" matching fragment attribute";

	catalog<String>("FragmentAttr")="mesh";
	catalog<String>("FragmentAttr","label")="Fragment Attribute";
	catalog<String>("FragmentAttr","enabler")="Bind Fragments";
	catalog<String>("FragmentAttr","hint")=
			"Name of string attribute describing membership"
			" in a collection of primitives";

	catalog<bool>("Bind Primitives")=false;
	catalog<bool>("Bind Primitives","joined")=true;
	catalog<String>("Bind Primitives","hint")=
			"Affect whole primitives instead of individual points";

	catalog<bool>("Conform")=false;
	catalog<String>("Conform","label")="Conform Primitives";
	catalog<String>("Conform","hint")=
			"Force all points of each input primitive to bind to"
			" the same subset of driver primitives."
			"  This can be useful to coerce all points on"
			" each input curve to follow the same driver curve(s).";

	catalog<I32>("Bindings")=1;
	catalog<I32>("Bindings","min")=1;
	catalog<I32>("Bindings","high")=100;
	catalog<I32>("Bindings","max")=1e3;
	catalog<bool>("Bindings","joined")=true;
	catalog<String>("Bindings","hint")=
			"Maximum number of attachment contacts.";

	catalog<bool>("Covert")=false;
	catalog<String>("Covert","label")="Covert Bindings";
	catalog<String>("Covert","hint")=
			"Store bindings apart from the mesh.";

	catalog<bool>("LimitDistance")=false;
	catalog<String>("LimitDistance","label")="Limit Distance";
	catalog<bool>("LimitDistance","joined")=true;
	catalog<String>("LimitDistance","hint")=
			"Dismiss potential bindings beyond a specific distance.";

	catalog<Real>("MaxDistance")=1.0;
	catalog<Real>("MaxDistance","high")=10.0;
	catalog<Real>("MaxDistance","max")=1e6;
	catalog<String>("MaxDistance","label")="To";
	catalog<String>("MaxDistance","enabler")="LimitDistance";
	catalog<String>("MaxDistance","hint")=
			"Disallow bindings beyond this distance.";

	catalog<bool>("PartitionDriver")=false;
	catalog<String>("PartitionDriver","label")="Partition Driver";
	catalog<bool>("PartitionDriver","joined")=true;
	catalog<String>("PartitionDriver","hint")=
			"Restrict bindings by matching attribute values.";

	catalog<String>("DriverPartitionAttr")="partition";
	catalog<String>("DriverPartitionAttr","label")="Driver Attr";
	catalog<String>("DriverPartitionAttr","enabler")="PartitionDriver";
	catalog<bool>("DriverPartitionAttr","joined")=true;
	catalog<String>("DriverPartitionAttr","hint")=
			"Primitive string attribute on driver to use for bind matching."
			"  Generally, these values are just simple names.";

	catalog<String>("InputPartitionAttr")="partition";
	catalog<String>("InputPartitionAttr","label")="Input Attr";
	catalog<String>("InputPartitionAttr","enabler")="PartitionDriver";
	catalog<String>("InputPartitionAttr","hint")=
			"Point string attribute on input to use for bind matching."
			"  Each value can be a simple name or a full regex pattern."
			"  For example, use \"prefix.*\" for every name starting"
			" with particular value"
			" (the '.' is required to indicate 'any character' and"
			" the '*' means 'repeated without limit')"
			" or \"one|two|three\" to allow any one name from a list."
			"  Be careful with non-letter characters that may have"
			" a special meaning in regular expressions.";

	catalog<String>("PivotPoint")="First";
	catalog<String>("PivotPoint","label")="Pivot Point";
	catalog<String>("PivotPoint","choice:0")="First";
	catalog<String>("PivotPoint","choice:1")="Center";
	catalog<String>("PivotPoint","choice:2")="Nearest";
	catalog<String>("PivotPoint","hint")=
			"Location used to determine attachment.";

	catalog< sp<Component> >("Input Surface");
	catalog<String>("Input Surface","label")="Reference Input Surface";

	catalog< sp<Component> >("Driver Surface");
	catalog<String>("Driver Surface","label")="Reference Driver Surface";
}

void SurfaceBindOp::handle(Record& a_rSignal)
{
#if FE_SBO_DEBUG
	feLog("SurfaceBindOp::handle\n");
#endif

	m_partitionDriver=catalog<bool>("PartitionDriver");
	m_bindPrimitives=catalog<bool>("Bind Primitives");
	m_conform=!m_bindPrimitives && catalog<bool>("Conform");

	const String fragmentAttribute=catalog<String>("FragmentAttr");
	const String driverPartitionAttr=catalog<String>("DriverPartitionAttr");
	const BWORD covert=catalog<bool>("Covert");

	BWORD warnedConformTriangle=FALSE;

	if(m_bindPrimitives && catalog<bool>("Conform"))
	{
		catalog<String>("warning")+=
				"Conforming only supports point bindings;";
	}

	//* output surface is copy of input surface
	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	m_element=m_bindPrimitives? e_primitive: e_point;
	m_attribute=m_bindPrimitives? e_vertices: e_position;

	const String covertKey=name();	//* TODO path also?

	m_spBindingAccessible=m_spOutputAccessible;
	if(covert)
	{
		sp<SurfaceAccessibleCatalog> spSurfaceAccessibleCatalog=
				registry()->create("*.SurfaceAccessibleCatalog");
		if(spSurfaceAccessibleCatalog.isValid())
		{
			spSurfaceAccessibleCatalog->setCatalog(
					registry()->master()->catalog());
			spSurfaceAccessibleCatalog->setKey(covertKey);
			m_spBindingAccessible=spSurfaceAccessibleCatalog;
		}
	}

	const String driverGroup=catalog<String>("DriverGroup");

	if(!access(m_spDriver,"Driver Surface",driverGroup)) return;

	sp<SurfaceAccessorI> spDriverNormals;
	if(!access(spDriverNormals,"Driver Surface",e_point,e_normal) ||
			!spDriverNormals->count())
	{
		catalog<String>("warning")+=
				"Driver Surface has no normals (expect rotational problems);";
	}
	spDriverNormals=NULL;

	if(m_partitionDriver)
	{
		m_spDriver->partitionWith(driverPartitionAttr);
	}

	m_spDriver->setRefinement(0);
	m_spDriver->prepareForSearch();

	if(m_spDriver->radius()<1e-6)
	{
		catalog<String>("error")=
				"Driver Surface appears empty (zero radius bounding sphere);";
		return;
	}

	sp<SurfaceAccessorI> spOutputVertices;
	if((m_conform || m_bindPrimitives) && !access(spOutputVertices,
			m_spOutputAccessible,e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spOutputElements;
	if(!access(spOutputElements,m_spOutputAccessible,
			m_element,m_attribute)) return;

	const BWORD bindFragments=catalog<bool>("Bind Fragments");
	m_fragmented=(bindFragments && spOutputElements.isValid());

	I32 fragmentCount=1;
	if(m_fragmented)
	{
		spOutputElements->fragmentWith(
				SurfaceAccessibleI::Element(m_element),fragmentAttribute);
		fragmentCount=spOutputElements->fragmentCount();
	}

	if(!access(m_spDriverAccessible,"Driver Surface")) return;

	std::vector< sp<SurfaceAccessorI> > outputFaces;
	std::vector< sp<SurfaceAccessorI> > outputBarys;
	const I32 bindingsAccessed=accessBindings(m_spBindingAccessible,
			outputFaces,outputBarys,
			catalog<I32>("Bindings"),m_bindPrimitives,TRUE);
	if(bindingsAccessed<1)
	{
		return;
	}

	const String signature=generateSignature(m_spDriverAccessible);
	storeSignature(m_spOutputAccessible,signature);

	storeParameters(m_spOutputAccessible,driverGroup,fragmentAttribute,
			bindingsAccessed,m_bindPrimitives,bindFragments,
			covert? covertKey: "");

	const U32 elementCount=spOutputElements->count();

	sp<SurfaceAccessorI> spDriverVertices;
	if(m_conform)
	{
		if(!access(spDriverVertices,m_spDriverAccessible,
				e_primitive,e_vertices)) return;

		//* default -1 = no master
		m_masterBinding.clear();
		m_masterBinding.resize(elementCount,-1);

		const U32 driverPrimitiveCount=spDriverVertices->count();
		m_driverFacePrimitive.resize(driverPrimitiveCount*2,-1);

		sp<SurfaceAccessorI> spOutputProperties;
		if(!access(spOutputProperties,m_spOutputAccessible,
				e_primitive,e_properties)) return;

		const U32 primitiveCount=spOutputVertices->count();
		for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
		{
			if(!spOutputProperties->integer(primitiveIndex,e_openCurve))
			{
				if(!warnedConformTriangle)
				{
					catalog<String>("warning")+="Conforming only supports"
							" open curves in Input Surface;";
					warnedConformTriangle=TRUE;
				}
				continue;
			}

			const U32 subCount=spOutputVertices->subCount(primitiveIndex);
			if(subCount<2)
			{
				continue;
			}

			const I32 masterIndex=spOutputVertices->integer(primitiveIndex,0);

			for(U32 subIndex=1;subIndex<subCount;subIndex++)
			{
				const I32 pointIndex=
						spOutputVertices->integer(primitiveIndex,subIndex);
				m_masterBinding[pointIndex]=masterIndex;
			}
		}
	}

	//* TODO fragment-aware ranges
	//* NOTE conform unsafe by using m_driverFacePrimitive table
	//* NOTE covert access to Catalog is not thread safe
	//* NOTE partition filter setting is per surface
	if(m_fragmented || m_conform || covert || m_partitionDriver)
	{
		//* serial
		sp<SurfaceRange> spFullRange=sp<SurfaceRange>(new SurfaceRange());
		spFullRange->nonAtomic().add(0,
				(m_fragmented? fragmentCount: elementCount)-1);
		run(-1,spFullRange);
	}
	else
	{
		setAtomicRange(m_spOutputAccessible,
				m_bindPrimitives? e_pointsOfPrimitives: e_pointsOnly);

		OperatorThreaded::handle(a_rSignal);
	}

#if FE_SBO_DEBUG
	feLog("SurfaceBindOp::handle done\n");
#endif
}

void SurfaceBindOp::run(I32 a_id,sp<SurfaceRange> a_spRange)
{
	const String fragmentAttribute=catalog<String>("FragmentAttr");
	const String inputPartitionAttr=catalog<String>("InputPartitionAttr");
	const String pivotPoint=catalog<String>("PivotPoint");
	const BWORD limitDistance=catalog<bool>("LimitDistance");
	const Real maxDistance=catalog<Real>("MaxDistance");

	std::vector< sp<SurfaceAccessorI> > outputFaces;
	std::vector< sp<SurfaceAccessorI> > outputBarys;
	const I32 bindingsAccessed=accessBindings(m_spBindingAccessible,
			outputFaces,outputBarys,
			catalog<I32>("Bindings"),m_bindPrimitives,TRUE);
	if(bindingsAccessed<1)
	{
		return;
	}

	sp<SurfaceAccessorI> spOutputElements;
	if(!access(spOutputElements,m_spOutputAccessible,
			m_element,m_attribute)) return;

	//* TODO don't redo fragmentation
	if(m_fragmented)
	{
		spOutputElements->fragmentWith(
				SurfaceAccessibleI::Element(m_element),fragmentAttribute);
	}

	sp<SurfaceAccessorI> spInputPartition;
	if(m_partitionDriver)
	{
		if(!access(spInputPartition,"Input Surface",
				m_element,inputPartitionAttr,e_error)) return;
	}

	sp<SurfaceAccessorI> spDriverVertices;
	sp<SurfaceAccessorI> spDriverProperties;
	if(m_conform)
	{
		if(!access(spDriverVertices,m_spDriverAccessible,
				e_primitive,e_vertices)) return;
		if(!access(spDriverProperties,m_spDriverAccessible,
				e_primitive,e_properties)) return;
	}

	I32 face= -1;
	Barycenter<Real> barycenter;

	sp<SurfaceAccessorI::FilterI> spFilter;

	for(SurfaceRange::Iterator it=a_spRange->begin();!it.atEnd();it.step())
	{
		const U32 index=it.value();

		I32 filterCount=1;
		if(m_fragmented)
		{
			const String fragment=spOutputElements->fragment(index);
			if(!spOutputElements->filterWith(fragment,spFilter) ||
					spFilter.isNull())
			{
				continue;
			}
			filterCount=spFilter->filterCount();
			if(!filterCount)
			{
				continue;
			}
		}

		if(m_partitionDriver)
		{
			const I32 reIndex=m_fragmented? spFilter->filter(0): index;

			const String partition=spInputPartition->string(reIndex);
			m_spDriver->setPartitionFilter(partition);

#if FE_SBO_VERBOSE
			feLog("SurfaceBindOp::handle element %d filter \"%s\"\n",
					index,partition.c_str());
#endif
		}

		SpatialVector pivot;

		if(pivotPoint=="Nearest")
		{
			set(pivot);
			Real bestDistance= -1.0;
			for(I32 filterIndex=0;filterIndex<filterCount;filterIndex++)
			{
				const I32 reIndex=m_fragmented?
						spFilter->filter(filterIndex): index;
				const U32 subCount=spOutputElements->subCount(reIndex);

				for(U32 subIndex=0;subIndex<subCount;subIndex++)
				{
					const SpatialVector point=
							spOutputElements->spatialVector(reIndex,subIndex);
					sp<SurfaceI::ImpactI> impact=
							m_spDriver->nearestPoint(point);
					const Real distance=impact->distance();
					if(bestDistance<0.0 || bestDistance>distance)
					{
						bestDistance=distance;
						pivot=point;
					}
				}
			}
		}
		else if(pivotPoint=="Center")
		{
			U32 sum=0;
			set(pivot);
			for(I32 filterIndex=0;filterIndex<filterCount;filterIndex++)
			{
				const I32 reIndex=m_fragmented?
						spFilter->filter(filterIndex): index;
				const U32 subCount=spOutputElements->subCount(reIndex);

				for(U32 subIndex=0;subIndex<subCount;subIndex++)
				{
					pivot+=spOutputElements->spatialVector(reIndex,subIndex);
				}
				sum+=subCount;
			}
			if(!sum)
			{
				continue;
			}
			pivot*=1.0/Real(sum);
		}
		else
		{
			const I32 reIndex=m_fragmented? spFilter->filter(0): index;
			if(!spOutputElements->subCount(reIndex))
			{
				continue;
			}
			pivot=spOutputElements->spatialVector(reIndex,0);
		}

		if(m_conform)
		{
			const I32 reIndex=m_fragmented? spFilter->filter(0): index;
			const I32 masterIndex=m_masterBinding[reIndex];
			if(masterIndex>=0)
			{
				for(U32 binding=0;binding<bindingsAccessed;binding++)
				{
					face=outputFaces[binding]->integer(masterIndex);
					outputFaces[binding]->set(reIndex,face);

					if(face<0)
					{
						continue;
					}

					const I32 driverPrimitiveIndex=m_driverFacePrimitive[face];

					if(driverPrimitiveIndex>=0 && spDriverProperties->integer(
							driverPrimitiveIndex,e_openCurve))
					{
						const U32 subCount=spDriverVertices->subCount(face);

						set(barycenter);
						barycenter[0]=subCount? nearestPointAlongCurve(pivot,
								spDriverVertices,face)/(subCount-1.0): 0.0;
						barycenter[1]=1.0-barycenter[0];
					}
					else
					{
						barycenter=outputBarys[binding]->spatialVector(
								masterIndex);
					}

					outputBarys[binding]->set(reIndex,barycenter);
				}

				continue;
			}
		}

		std::vector< sp<SurfaceI::ImpactI> > impactArray=
				m_spDriver->nearestPoints(pivot,bindingsAccessed);
		const U32 impacts=impactArray.size();

		for(U32 binding=0;binding<bindingsAccessed;binding++)
		{
			face= -1;
			set(barycenter);

			if(binding<impacts && impactArray[binding].isValid())
			{
				sp<SurfaceI::ImpactI>& rspImpact=impactArray[binding];
				const Real distance=rspImpact->distance();
				if(!limitDistance || distance<=maxDistance)
				{
					face=rspImpact->triangleIndex();
					barycenter=rspImpact->barycenter();
					if(m_conform)
					{
						FEASSERT(face<m_driverFacePrimitive.size());
						m_driverFacePrimitive[face]=rspImpact->primitiveIndex();
					}
				}
			}

			if(m_fragmented)
			{
				for(I32 filterIndex=0;filterIndex<filterCount;filterIndex++)
				{
					const I32 filtered=spFilter->filter(filterIndex);

					outputFaces[binding]->set(filtered,face);
					outputBarys[binding]->set(filtered,barycenter);
				}
			}
			else
			{
				outputFaces[binding]->set(index,face);
				outputBarys[binding]->set(index,barycenter);
			}
#if FE_SBO_VERBOSE
			feLog("SurfaceBindOp::handle element %d bind %d/%d"
					" pivot %s face %d bary %s\n",
					index,binding,bindingsAccessed,
					c_print(pivot),face,c_print(barycenter));
#endif
		}
	}
}

String SurfaceBindOp::retrieveSignature(
	sp<SurfaceAccessibleI>& a_rspAccessibleI)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	String signatureName;
	signatureName.sPrintf("%sSignature",bindPrefix.c_str());

	sp<SurfaceAccessorI> spSignatureAccessor;
	if(!access(spSignatureAccessor,a_rspAccessibleI,
			e_detail,signatureName,e_warning))
	{
		return "";
	}

	return spSignatureAccessor->string(0);
}

void SurfaceBindOp::storeSignature(sp<SurfaceAccessibleI>& a_rspAccessibleI,
	String a_signature)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	String signatureName;
	signatureName.sPrintf("%sSignature",bindPrefix.c_str());

	sp<SurfaceAccessorI> spSignatureAccessor;
	if(!access(spSignatureAccessor,a_rspAccessibleI,
			e_detail,signatureName,e_error,e_createMissing)) return;

	spSignatureAccessor->set(0,a_signature);

}

void SurfaceBindOp::discardSignature(sp<SurfaceAccessibleI>& a_rspAccessibleI)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	String signatureName;
	signatureName.sPrintf("%sSignature",bindPrefix.c_str());

	discard(a_rspAccessibleI,e_detail,signatureName);
}

void SurfaceBindOp::retrieveParameters(sp<SurfaceAccessibleI>& a_rspAccessibleI,
	String& a_rDriverGroup,String& a_rFragmentAttribute,
	I32& a_rBindings,BWORD& a_rBindPrimitives,BWORD& a_rBindFragments,
	String& a_rCovertKey)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	String attributeName;
	sp<SurfaceAccessorI> spAccessor;

	attributeName.sPrintf("%sDriverGroup",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rDriverGroup=spAccessor->string(0);
	}

	attributeName.sPrintf("%sFragment",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rFragmentAttribute=spAccessor->string(0);
	}

	attributeName.sPrintf("%sCount",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rBindings=spAccessor->integer(0);
	}

	attributeName.sPrintf("%sPrimitives",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rBindPrimitives=spAccessor->integer(0);
	}

	attributeName.sPrintf("%sFragments",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rBindFragments=spAccessor->integer(0);
	}

	attributeName.sPrintf("%sCovert",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning))
	{
		a_rCovertKey=spAccessor->string(0);
	}
}

void SurfaceBindOp::storeParameters(sp<SurfaceAccessibleI>& a_rspAccessibleI,
	String a_driverGroup,String a_fragmentAttribute,
	I32 a_bindCount,BWORD a_bindPrimitives,BWORD a_bindFragments,
	String a_covertKey)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	String attributeName;
	sp<SurfaceAccessorI> spAccessor;

	attributeName.sPrintf("%sDriverGroup",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_driverGroup);
	}

	attributeName.sPrintf("%sFragment",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_fragmentAttribute);
	}

	attributeName.sPrintf("%sCount",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_bindCount);
	}

	attributeName.sPrintf("%sPrimitives",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_bindPrimitives);
	}

	attributeName.sPrintf("%sFragments",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_bindFragments);
	}

	attributeName.sPrintf("%sCovert",bindPrefix.c_str());
	if(access(spAccessor,a_rspAccessibleI,
			e_detail,attributeName,e_warning,e_createMissing))
	{
		spAccessor->set(0,a_covertKey);
	}
}

void SurfaceBindOp::discardParameters(sp<SurfaceAccessibleI>& a_rspAccessibleI)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	String attributeName;
	sp<SurfaceAccessorI> spAccessor;

	attributeName.sPrintf("%sDriverGroup",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);

	attributeName.sPrintf("%sFragment",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);

	attributeName.sPrintf("%sCount",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);

	attributeName.sPrintf("%sPrimitives",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);

	attributeName.sPrintf("%sFragments",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);

	attributeName.sPrintf("%sCovert",bindPrefix.c_str());
	discard(a_rspAccessibleI,e_detail,attributeName);
}

I32 SurfaceBindOp::accessBindings(sp<SurfaceAccessibleI>& a_rspAccessibleI,
	std::vector< sp<SurfaceAccessorI> >&a_rFaces,
	std::vector< sp<SurfaceAccessorI> >&a_rBarys,
	I32 a_bindCount,BWORD a_bindPrimitives,BWORD a_allowCreation)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	const SurfaceAccessibleI::Element element=a_bindPrimitives?
			SurfaceAccessibleI::e_primitive:
			SurfaceAccessibleI::e_point;

	const SurfaceAccessibleI::Creation create=a_allowCreation?
			SurfaceAccessibleI::e_createMissing:
			SurfaceAccessibleI::e_refuseMissing;

	a_rFaces.resize(a_bindCount);
	a_rBarys.resize(a_bindCount);
	U32 bindingsAccessed=0;
	for(U32 binding=0;binding<a_bindCount;binding++)
	{
		String faceName;
		String baryName;
		faceName.sPrintf("%sFace%d",bindPrefix.c_str(),binding);
		baryName.sPrintf("%sBary%d",bindPrefix.c_str(),binding);

		a_rFaces[binding]=a_rspAccessibleI->accessor(element,faceName,create);
		if(a_rFaces[binding].isNull())
		{
			if(a_allowCreation)
			{
				feLog("SurfaceBindOp face attribute"
						" \"%s\" not accessible\n",faceName.c_str());
				catalog<String>("error")=
						"inaccessible face attribute \""+faceName+"\"";
				return -1;
			}
			break;
		}

		a_rBarys[binding]=a_rspAccessibleI->accessor(element,baryName,create);
		if(a_rBarys[binding].isNull())
		{
			if(a_allowCreation)
			{
				feLog("SurfaceBindOp barycentric attribute"
						" \"%s\" not accessible\n",baryName.c_str());
				catalog<String>("error")="inaccessible barycentric"
						" attribute \""+baryName+"\"";
				return -1;
			}
			break;
		}

		//* force creation now
		if(a_allowCreation)
		{
			a_rFaces[binding]->set(0,0);
			a_rBarys[binding]->set(0,Barycenter<Real>(0.0,0.0));
		}

		bindingsAccessed++;
	}

	return bindingsAccessed;
}

void SurfaceBindOp::discardBindings(sp<SurfaceAccessibleI>& a_rspAccessibleI,
	I32 a_bindCount,BWORD a_bindPrimitives)
{
	const String bindPrefix=catalog<String>("BindPrefix");

	const Element element=a_bindPrimitives? e_primitive: e_point;

	for(U32 binding=0;binding<a_bindCount;binding++)
	{
		String faceName;
		String baryName;
		faceName.sPrintf("%sFace%d",bindPrefix.c_str(),binding);
		baryName.sPrintf("%sBary%d",bindPrefix.c_str(),binding);

		discard(a_rspAccessibleI,element,faceName);
		discard(a_rspAccessibleI,element,baryName);
	}
}
